/*
* Copyright(c) 1997-2001 Id Software, Inc.
* Copyright(c) 2002 The Quakeforge Project.
* Copyright(c) 2006 Quetoo.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or(at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
* 
* See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <dirent.h>

#include "g_local.h"

game_locals_t game;
level_locals_t level;
game_import_t gi;
game_export_t globals;
spawn_temp_t st;

int sm_meat_index;
int snd_fry;
int meansOfDeath;

edict_t *g_edicts;

cvar_t *dmflags;
cvar_t *fraglimit;
cvar_t *timelimit;
cvar_t *password;
cvar_t *spectator_password;
cvar_t *needpass;
cvar_t *maxclients;
cvar_t *maxspectators;
cvar_t *maxentities;
cvar_t *g_select_empty;
cvar_t *dedicated;

cvar_t *filterban;

cvar_t *sv_maxvelocity;
cvar_t *sv_gravity;

cvar_t *sv_rollspeed;
cvar_t *sv_rollangle;
cvar_t *gun_x;
cvar_t *gun_y;
cvar_t *gun_z;

cvar_t *run_pitch;
cvar_t *run_roll;
cvar_t *bob_up;
cvar_t *bob_pitch;
cvar_t *bob_roll;

cvar_t *sv_cheats;

cvar_t *flood_msgs;
cvar_t *flood_persecond;
cvar_t *flood_waitdelay;

cvar_t *gameplay;

cvar_t *hook;

team_t good = { "Good", "male/howitzer", 0, 0.0, 0.0},
	evil = { "Evil", "female/jezebel", 0, 0.0, 0.0};

cvar_t *teams;

cvar_t *sv_allow_vote;

cvar_t *sv_inter_sound;

cvar_t *sv_maplist;
cvar_t *sv_rmap;
maplist_t maplist;

cvar_t *log_fraglog;
cvar_t *log_chatlog;

#ifdef HAVE_MYSQL
cvar_t *log_mysql;
cvar_t *mysql_host;
cvar_t *mysql_user;
cvar_t *mysql_password;
cvar_t *mysql_db;
MYSQL *mysql;
char sql[512];
#endif

char *q2home;
char *home;

FILE *fraglog, *chatlog, *f;

void SpawnEntities(char *mapname, char *entities, char *spawnpoint);
void ClientThink(edict_t *ent, usercmd_t *cmd);
qboolean ClientConnect(edict_t *ent, char *userinfo);
void ClientUserinfoChanged(edict_t *ent, char *userinfo);
void ClientDisconnect(edict_t *ent);
void ClientBegin(edict_t *ent);
void ClientCommand(edict_t *ent);
void RunEntity(edict_t *ent);
void ShutdownGame(void);
void InitGame(void);
void InitGameFS(void);
FILE *OpenFile(char *filename, char *mode);
void G_RunFrame(void);


/*
InitGame
 
This will be called when the dll is first loaded, which
only happens when a new game is started or a save game
is loaded.
*/
void InitGame(void){
	gi.dprintf("==== InitGame ====\n");
	
	gun_x = gi.cvar("gun_x", "0", 0);
	gun_y = gi.cvar("gun_y", "0", 0);
	gun_z = gi.cvar("gun_z", "0", 0);
	
	//FIXME: sv_ prefix is wrong for these
	sv_rollspeed = gi.cvar("sv_rollspeed", "200", 0);
	sv_rollangle = gi.cvar("sv_rollangle", "2", 0);
	sv_maxvelocity = gi.cvar("sv_maxvelocity", "2000", 0);
	sv_gravity = gi.cvar("sv_gravity", "800", 0);
	
	// noset vars
	dedicated = gi.cvar("dedicated", "0", CVAR_NOSET);
	
	// latched vars
	sv_cheats = gi.cvar("cheats", "0", CVAR_SERVERINFO | CVAR_LATCH);
	gi.cvar("gamename", GAMEVERSION , CVAR_SERVERINFO | CVAR_LATCH);
	gi.cvar("gamedate", __DATE__ , CVAR_SERVERINFO | CVAR_LATCH);
	
	maxclients = gi.cvar("maxclients", "4", CVAR_SERVERINFO | CVAR_LATCH);
	maxspectators = gi.cvar("maxspectators", "4", CVAR_SERVERINFO);
	maxentities = gi.cvar("maxentities", "1024", CVAR_LATCH);
	
	// change anytime vars
	dmflags = gi.cvar("dmflags", "0", CVAR_SERVERINFO);
	fraglimit = gi.cvar("fraglimit", "0", CVAR_SERVERINFO);
	timelimit = gi.cvar("timelimit", "0", CVAR_SERVERINFO);
	password = gi.cvar("password", "", CVAR_USERINFO);
	spectator_password = gi.cvar("spectator_password", "", CVAR_USERINFO);
	needpass = gi.cvar("needpass", "0", CVAR_SERVERINFO);
	filterban = gi.cvar("filterban", "1", 0);
	
	g_select_empty = gi.cvar("g_select_empty", "0", CVAR_ARCHIVE);
	
	run_pitch = gi.cvar("run_pitch", "0.002", 0);
	run_roll = gi.cvar("run_roll", "0.005", 0);
	bob_up = gi.cvar("bob_up", "0.005", 0);
	bob_pitch = gi.cvar("bob_pitch", "0.002", 0);
	bob_roll = gi.cvar("bob_roll", "0.002", 0);
	
	// flood control
	flood_msgs = gi.cvar("flood_msgs", "4", 0);
	flood_persecond = gi.cvar("flood_persecond", "4", 0);
	flood_waitdelay = gi.cvar("flood_waitdelay", "10", 0);
	
	gameplay = gi.cvar("gameplay", "0", 0);  // dm, instagib, arena
	
	hook = gi.cvar("hook", "0", 0);  // hook
	
	teams = gi.cvar("teams", "0", 0);  // teams
	
	sv_allow_vote = gi.cvar("sv_allow_vote", "1", 0);  // voting
	
	sv_inter_sound = gi.cvar("sv_inter_sound", "qmass.wav", 0);
	
	InitGameFS();  // init filesystem
	
	// dm map list, from maps.lst file
	sv_maplist = gi.cvar("sv_maplist", "maps.lst", 0);
	if(sv_maplist->string[0] && (f = OpenFile(sv_maplist->string, "r"))){
		int i = 0;
		
		while(fgets(maplist.maps[i], MAX_STRING_CHARS, f))
			if(++i == 64) break;  // load maps.lst
		
		maplist.count = i;
		maplist.index = 0;
		
		fclose(f);
	}
	else {
		gi.dprintf("Couldn't open %s\n", sv_maplist->string);
		maplist.count = 0;
	}
	
	// random map enabled/disabled
	sv_rmap = gi.cvar("sv_rmap", "0", 0);
	
	log_fraglog = gi.cvar("log_fraglog", "1", 0);
	log_chatlog = gi.cvar("log_chatlog", "1", 0);
	
	if(log_fraglog->value)
		fraglog = OpenFile("fraglog.log", "a");
		
	if(log_chatlog->value)
		chatlog = OpenFile("chatlog.log", "a");
		
#ifdef HAVE_MYSQL
	log_mysql = gi.cvar("log_mysql", "0", 0);
	
	mysql_host = gi.cvar("mysql_host", "localhost", 0);
	mysql_user = gi.cvar("mysql_user", "quake2", 0);
	mysql_password = gi.cvar("mysql_password", "", 0);
	mysql_db = gi.cvar("mysql_db", "quake2", 0);
	
	if(log_mysql->value){  //init database
	
		mysql = mysql_init(NULL);
		
		mysql_real_connect(mysql, mysql_host->string,
				mysql_user->string, mysql_password->string,
				mysql_db->string, 0, NULL, 0
		);
	}
#endif
	
	InitItems();
	
	game.helpmessage1[0] = 0;
	game.helpmessage2[0] = 0;
	
	// initialize all entities for this game
	g_edicts = gi.TagMalloc(maxentities->value * sizeof(g_edicts[0]), TAG_GAME);
	globals.edicts = g_edicts;
	globals.max_edicts = maxentities->value;
	
	// initialize all clients for this game
	maxclients->value = maxclients->value;
	game.clients = gi.TagMalloc(maxclients->value * sizeof(game.clients[0]), TAG_GAME);
	globals.num_edicts = maxclients->value + 1;
}

/*
*  Initializes the mod filesystem, rooted to QUAKE2_HOME/qmass
*  or HOME/.quake2/qmass.  The frag and chat logs are also opened here.
*/
void InitGameFS(void){

	//resolve q2home, home from environment
	q2home = getenv("QUAKE2_HOME");
	home = getenv("HOME");
	
	gi.dprintf("using %s as q2home\n", q2home);
	gi.dprintf("using %s as home\n", home);
}

/*
*  Tries to open the specified filename in q2home or home.
*/
FILE *OpenFile(char *filename, char *mode){

	FILE *f = NULL;
	
	if(q2home != NULL)  //open in q2home/qmass
		f = fopen(va("%s/qmass/%s", q2home, filename), mode);
		
	if(f == NULL && home != NULL)  //or in ~/.quake2/qmass
		f = fopen(va("%s/.quake2/qmass/%s", home, filename), mode);
		
	if(f == NULL)  //or wherever we are
		f = fopen(filename, mode);
		
	if(f != NULL)
		gi.dprintf("opened %s with mode %s\n", filename, mode);
		
	return f;
}

/*
*  Frees tags and closes fraglog.  This is called when the 
*  game is unloaded (complements InitGame).
*/
void ShutdownGame(void){
	gi.dprintf("==== ShutdownGame ====\n");
	
	if(fraglog != NULL)
		fclose(fraglog);  //close fraglog
		
	if(chatlog != NULL)
		fclose(chatlog);  //and chatlog
	
#ifdef HAVE_MYSQL
	if(mysql != NULL)
		mysql_close(mysql);  //and db
#endif
	
	gi.FreeTags(TAG_LEVEL);
	gi.FreeTags(TAG_GAME);
}

/*
We stub out some things that other engines will require so that our mods
can be run on them.
*/


void WriteGame(char *filename, qboolean autosave){}
void ReadGame(char *filename){}
void WriteLevel(char *filename){}
void ReadLevel(char *filename){}


/*
GetGameAPI
 
Returns a pointer to the structure with all entry points
and global variables
*/
game_export_t *GetGameAPI(game_import_t *import){
	gi = *import;
	
	globals.apiversion = GAME_API_VERSION;
	globals.Init = InitGame;
	globals.Shutdown = ShutdownGame;
	globals.SpawnEntities = SpawnEntities;
	
	globals.WriteGame = WriteGame;
	globals.ReadGame = ReadGame;
	globals.WriteLevel = WriteLevel;
	globals.ReadLevel = ReadLevel;
	
	globals.ClientThink = ClientThink;
	globals.ClientConnect = ClientConnect;
	globals.ClientUserinfoChanged = ClientUserinfoChanged;
	globals.ClientDisconnect = ClientDisconnect;
	globals.ClientBegin = ClientBegin;
	globals.ClientCommand = ClientCommand;
	
	globals.RunFrame = G_RunFrame;
	
	globals.ServerCommand = ServerCommand;
	
	globals.edict_size = sizeof(edict_t);
	
	return &globals;
}

/*
Com_Printf

Redefined here so functions in q_shared.c can link.
*/
void Com_Printf(char *msg, ...){
	va_list	argptr;
	char text[1024];
	
	va_start(argptr, msg);
	vsprintf(text, msg, argptr);
	va_end(argptr);
	
	gi.dprintf("%s", text);
}

/*
ClientEndServerFrames
*/
void ClientEndServerFrames(void){
	int i;
	edict_t *ent;
	
	// calc the player views now that all pushing
	// and damage has been added
	for(i = 0; i < maxclients->value; i++){
		ent = g_edicts + 1 + i;
		if(!ent->inuse || !ent->client)
			continue;
		ClientEndServerFrame(ent);
	}
}

/*
CreateTargetChangeLevel
 
Returns the created target changelevel
*/
edict_t *CreateTargetChangeLevel(char *map){
	edict_t *ent;
	
	ent = G_Spawn();
	ent->classname = "target_changelevel";
	Com_sprintf(level.nextmap, sizeof(level.nextmap), "%s", map);
	ent->map = level.nextmap;
	return ent;
}


/*
EndDMLevel
 
The timelimit or fraglimit has been exceeded
*/
void EndDMLevel(void){
	char *map;
	
	// stay on same level flag
	if((int)dmflags->value & DF_SAME_LEVEL){
		BeginIntermission(CreateTargetChangeLevel(level.mapname));
		return;
	}
	
	// or try maplist
	if(maplist.count > 0){
		
		if(sv_rmap->value)
			maplist.index = rand() % maplist.count;
		else maplist.index = (maplist.index + 1) % maplist.count;
		
		map = strtok(maplist.maps[maplist.index], " ");
		BeginIntermission(CreateTargetChangeLevel(map));
		return;
	}
	
	if(!level.nextmap[0])  // or just restart current map
		strncpy(level.nextmap, level.mapname, MAX_QPATH);
	
	BeginIntermission(CreateTargetChangeLevel(level.nextmap));
}


/*
CheckNeedPass
*/
void CheckNeedPass(void){
	int need;
	
	// if password or spectator_password has changed, update needpass
	// as needed
	if(password->modified || spectator_password->modified){
		password->modified = spectator_password->modified = false;
		
		need = 0;
		
		if(*password->string && Q_stricmp(password->string, "none"))
			need |= 1;
		if(*spectator_password->string && Q_stricmp(spectator_password->string, "none"))
			need |= 2;
			
		gi.cvar_set("needpass", va("%d", need));
	}
}

/*
ResetVote
*/
void ResetVote(void){
	int i;
	
	for(i = 0; i < maxclients->value; i++){  //reset vote flags
		if(!g_edicts[i + 1].inuse)
			continue;
		g_edicts[i + 1].client->pers.vote = VOTE_NOOP;
	}
	
	votes[0] = votes[1] = votes[2] = 0;
	vote_active = false;
	vote_cmd[0] = 0;
}

/*
ResetScores
*/
void ResetScores(void){
	int i;
	
	good.score = evil.score = 0;
	
	for(i = 0; i < maxclients->value; i++)
		g_edicts[i + 1].client->resp.score = 0;
}

/*
CheckVote
*/
void CheckVote(void){
	int i, count = 0;
	
	if(!sv_allow_vote->value)
		return;
	
	if(!vote_active)
		return;
	
	if(level.time - vote_time > MAX_VOTE_TIME){
		gi.bprintf(PRINT_HIGH, "Vote \"%s\" expired\n", vote_cmd);
		ResetVote();
		return;
	}
	
	for(i = 0; i < maxclients->value; i++){
		if(!g_edicts[i + 1].inuse)
			continue;
		count++;
	}
	
	if(votes[VOTE_YES] >= count * VOTE_MAJORITY){  //vote passed
		gi.bprintf(PRINT_HIGH, "Vote \"%s\" passed\n", vote_cmd);
		if(strncmp(vote_cmd, "map ", 4) == 0)  //special case for map
			BeginIntermission(CreateTargetChangeLevel(vote_cmd + 4));
		else if(strcmp(vote_cmd, "reset") == 0)  //and reset
			ResetScores();
		else gi.AddCommandString(vote_cmd);
		ResetVote();
	} else if(votes[VOTE_NO] >= count * VOTE_MAJORITY){  //vote failed
		gi.bprintf(PRINT_HIGH, "Vote \"%s\" failed\n", vote_cmd);
		ResetVote();
	}
}

/*
CheckDMRules
*/
void CheckDMRules(void){
	int i;
	gclient_t *cl;
	char c[32];
	
	if(level.intermissiontime)
		return;
	
	if(timelimit->value){
		if(level.time >= timelimit->value * 60){
			gi.bprintf(PRINT_HIGH, "Timelimit hit.\n");
			EndDMLevel();
			return;
		}
	}
	
	if(fraglimit->value){
		
		if(teams->value){  //check team scores
 			if(good.score >= fraglimit->value ||
					evil.score >= fraglimit->value){
				gi.bprintf(PRINT_HIGH, "Fraglimit hit.\n");
				EndDMLevel();
				return;
			}
		}
		else {  //or individual scores
			for(i = 0; i < maxclients->value; i++){
				cl = game.clients + i;
				if(!g_edicts[i + 1].inuse)
					continue;
					
				if(cl->resp.score >= fraglimit->value){
					gi.bprintf(PRINT_HIGH, "Fraglimit hit.\n");
					EndDMLevel();
					return;
				}
			}
		}
	}
	
	if(gameplay->modified){  //change gameplay, fix items, respawn clients
		gameplay->modified = false;
		
		if(gameplay->value > ROCKET_ARENA || gameplay->value < DEATHMATCH)
			gameplay->value = DEATHMATCH;  //ensure a sane value
		
		for(i = 1; i < globals.num_edicts; i++){
			if(!g_edicts[i].inuse)
				continue;
			
			if(g_edicts[i].client)  //respawn clients
				PutClientInServer(&g_edicts[i]);
			else if(g_edicts[i].item){  //and toggle items
				if(gameplay->value){  //remove dropped, hide regular items
					if(g_edicts[i].spawnflags & DROPPED_ITEM){
						G_FreeEdict(&g_edicts[i]);
						continue;
					}
					g_edicts[i].svflags |= SVF_NOCLIENT;
					g_edicts[i].solid = SOLID_NOT;
					g_edicts[i].nextthink = 0;
				} else {  //unhide regular items
					g_edicts[i].svflags &= ~SVF_NOCLIENT;
					g_edicts[i].solid = SOLID_TRIGGER;
					g_edicts[i].nextthink = level.time + 0.2;
				}
			}
		}
		
		switch((int)gameplay->value){
			case INSTAGIB:
				strcpy(c, "INSTAGIB");
				break;
			case ROCKET_ARENA:
				strcpy(c, "ROCKET ARENA");
				break;
			default:
				strcpy(c, "DEATHMATCH");
				break;
		}
		gi.bprintf(PRINT_HIGH, "Gameplay has changed to %s!\n", c);
	}
	
	if(hook->modified){
		hook->modified = false;
		
		if(hook->value){  //stuff hook cmds
			gi.WriteByte(svc_stufftext);
			gi.WriteString("alias +hook \"cmd hook\"\n");
			gi.multicast(vec3_origin, MULTICAST_ALL_R);
			
			gi.WriteByte(svc_stufftext);
			gi.WriteString("alias -hook \"cmd unhook\"\n");
			gi.multicast(vec3_origin, MULTICAST_ALL_R);
		}
	}
	
	if(teams->modified){  //reset teams, make spectators
		teams->modified = false;
		
		for(i = 0; i < maxclients->value; i++){
			if(!g_edicts[i + 1].inuse)
				continue;
			
			cl = game.clients + i;
			
			cl->pers.spectator = false;
			cl->pers.team = NULL;
			
			PutClientInServer(&g_edicts[i + 1]);
			
			if(!teams->value){  //re-randomize male skin
				extern char *male_skins[];
				extern int male_skins_count;
				
				snprintf(cl->pers.skin, 16, 
					"male/%s", male_skins[rand() % male_skins_count]
				);
				
				gi.configstring(CS_PLAYERSKINS + i, 
					va("%s\\%s", cl->pers.netname, cl->pers.skin)
				);
			}
		}
		
		strcpy(c, teams->value ? "enabled" : "disabled");
		gi.bprintf(PRINT_HIGH, "Teams have been %s!\n", c);
	}
}


/*
ExitLevel
*/
void ExitLevel(void){
	int i;
	edict_t *ent;
	char command [256];
	
	Com_sprintf(command, sizeof(command), "gamemap \"%s\"\n", level.changemap);
	gi.AddCommandString(command);
	level.changemap = NULL;
	level.exitintermission = 0;
	level.intermissiontime = 0;
	ClientEndServerFrames();
	
	// clear some things before going to next level
	for(i = 0; i < maxclients->value; i++){
		ent = g_edicts + 1 + i;
		if(!ent->inuse)
			continue;
		if(ent->health > ent->client->pers.max_health)
			ent->health = ent->client->pers.max_health;
	}
	
	ResetScores();  //reset team scores
	ResetVote();  //make sure votes dont carry over
}

/*
G_RunFrame
 
Advances the world by 0.1 seconds
*/
void G_RunFrame(void){
	int i;
	edict_t *ent;
	
	level.framenum++;
	level.time = level.framenum * FRAMETIME;
	
	// exit intermissions
	
	if(level.exitintermission){
		ExitLevel();
		return;
	}
	
	// treat each object in turn
	// even the world gets a chance to think
	ent = &g_edicts[0];
	for(i = 0; i < globals.num_edicts; i++, ent++){
		if(!ent->inuse)
			continue;
			
		level.current_entity = ent;
		
		VectorCopy(ent->s.origin, ent->s.old_origin);
		
		// if the ground entity moved, make sure we are still on it
		if((ent->groundentity) && (ent->groundentity->linkcount != ent->groundentity_linkcount)){
			ent->groundentity = NULL;
		}
		
		if(i > 0 && i <= maxclients->value){
			ClientBeginServerFrame(ent);
			continue;
		}
		
		G_RunEntity(ent);
	}
	
	// see if a vote has passed
	CheckVote();
	
	// see if it is time to end a deathmatch
	CheckDMRules();
	
	// see if needpass needs updated
	CheckNeedPass();
	
	// build the playerstate_t structures for all players
	ClientEndServerFrames();
}

